
#include <triangulation/dim3.h>
#include <cstring>
#include <map>
#include <queue>
#include <vector>

#define MAXLEN 1000 // TODO: Work out what the real upper bound should be.
#define KILLSWITCH 5000000
// #define KILLSWITCH 5000

bool ALLOW_32;
bool ALLOW_20;

// This requires us to start with *everything* at a given level of the census.

using namespace regina;

struct Data;
typedef std::map<std::string, Data*> Graph;
typedef Graph::iterator GraphPtr;

Graph graph;
unsigned nComp;
std::queue<GraphPtr> q;
std::queue<GraphPtr> top;

unsigned maxTet;
unsigned long nBuild, nDescent;

struct Data {
    Data* parent;
    int depth;

    Data() : parent(0), depth(0) {}
};

Data* root(Data* n) {
    Data* ans = n;
    while (ans->parent)
        ans = ans->parent;
    // Compress the tree.
    if (n == ans)
        return ans;
    Data* tmp;
    while (n->parent != ans) {
        tmp = n->parent;
        n->parent = ans;
        n = tmp;
    }
    return ans;
}

// Return value is true iff we joined two distinct components.
bool join(Data* a, Data* b) {
    Data* aRoot = root(a);
    Data* bRoot = root(b);
    if (aRoot == bRoot)
        return false;

    if (aRoot->depth > bRoot->depth)
        bRoot->parent = aRoot;
    else if (bRoot->depth > aRoot->depth)
        aRoot->parent = bRoot;
    else {
        aRoot->parent = bRoot;
        ++bRoot->depth;
    }
    return true;
}

void init() {
    nComp = 0;
}

void cleanup() {
    for (auto elt : graph) {
        delete elt.second;
    }
    graph.clear();

    {
        std::queue<GraphPtr> empty;
        q.swap(empty);
    }
    {
        std::queue<GraphPtr> empty;
        top.swap(empty);
    }
}

void process(const std::string sig, Data* dat) {
    Triangulation<3>* t = Triangulation<3>::fromIsoSig(sig);
    if (! t) {
        std::cerr << "BAD ISOSIG: " << sig << std::endl;
    }

    int i;
    std::string next;
    GraphPtr pos;

    if (t->size() < maxTet) {
        for (i = 0; i < t->countTriangles(); ++i) {
            Triangulation<3> copy(*t, false);
            if (copy.pachner(copy.triangle(i), true, true)) {
                next = copy.isoSig();

                pos = graph.find(next);
                if (pos == graph.end()) {
                    pos = graph.insert(Graph::value_type(next, new Data())).first;
                    q.push(pos);
                    if (copy.size() == maxTet)
                        top.push(pos);
                    if (! join(dat, pos->second))
                        std::cerr << "ERROR: Adjacency problem!" << std::endl;
                    ++nBuild;
                } else {
                    if (join(dat, pos->second))
                        --nComp;
                }
            }
        }
    }

    if (ALLOW_32) {
        for (i = 0; i < t->countEdges(); ++i) {
            Triangulation<3> copy(*t, false);
            if (copy.pachner(copy.edge(i), true, true)) {
                next = copy.isoSig();

                pos = graph.find(next);
                if (pos == graph.end()) {
                    pos = graph.insert(Graph::value_type(next, new Data())).first;
                    q.push(pos);
                    if (! join(dat, pos->second))
                        std::cerr << "ERROR: Adjacency problem!" << std::endl;
                    ++nBuild;
                } else {
                    if (join(dat, pos->second))
                        --nComp;
                }
            }
        }
    }

    if (ALLOW_20) {
        std::queue<Triangulation<3>*> descent;
        std::set<std::string> seen;
        descent.push(t);

        while (! descent.empty()) {
            t = descent.front();
            descent.pop();

            for (i = 0; i < t->countEdges(); ++i) {
                if (t->twoZeroMove(t->edge(i), true, false)) {
                    Triangulation<3>* copy = new Triangulation<3>(*t, false);
                    copy->twoZeroMove(copy->edge(i), false, true);
                    next = copy->isoSig();
                    if (! seen.insert(next).second) {
                        delete copy;
                        continue;
                    }

                    descent.push(copy);
                    ++nDescent;

                    // pthread_mutex_lock(&mut);
                    pos = graph.find(next);
                    if (pos != graph.end()) {
                        if (join(dat, pos->second))
                            --nComp;
                    }
                    // TODO: Allow early break here
                }
            }

            delete t;
        }
    } else {
        delete t;
    }
}

void run(const std::string& manifold, const std::vector<std::string>& sigs,
        bool allow32, bool allow20) {
    ALLOW_32 = allow32;
    ALLOW_20 = allow20;

    init();

    nBuild = nDescent = 0;

    for (const auto& sig : sigs) {
        GraphPtr pos = graph.insert(Graph::value_type(sig, new Data)).first;
        top.push(pos);
        ++nComp;
    }

    int height = 0;
    while (nComp > 1 && nBuild < KILLSWITCH) {
        // std::cout << nComp << " components" << std::endl;
        ++height;
        maxTet = (top.front()->first[0] - 'a') + 1;
        top.swap(q);

        GraphPtr pos;
        while (nComp > 1 && (! q.empty()) && nBuild < KILLSWITCH) {
            pos = q.front();
            q.pop();
            process(pos->first, pos->second);
        }
    }

    cleanup();

    if (allow32) {
        if (allow20) {
            std::cout << "all";
        } else {
            std::cout << "zigzag";
        }
    } else {
        if (allow20) {
            std::cout << "semi-monotonic";
        } else {
            std::cout << "monotonic";
        }
    }
    if (nComp > 1) {
        if (nBuild >= KILLSWITCH)
            nBuild = 0;
        else
            std::cerr << "ERROR: Premature death" << std::endl;
    }
    std::cout << " [" << manifold << "] -> height,build,descent = "
        << height << ',' << nBuild << ',' << nDescent << std::endl;
}

int main() {
    std::string manifold;
    std::string input;
    std::vector<std::string> sigs;

    std::cin >> manifold;
    while (true) {
        std::cin >> input;
        if (std::cin.eof())
            break;

        regina::Triangulation<3>* tri =
            regina::Triangulation<3>::fromIsoSig(input);
        if (! tri) {
            std::cerr << "ERROR: " << input << std::endl;
            return 1;
        }
        sigs.push_back(tri->isoSig());
        delete tri;
    }

    run(manifold, sigs, true, false);
    run(manifold, sigs, false, false);
    run(manifold, sigs, false, true);

    return 0;
}

